use std::{fmt::Write, str::Chars};

#[derive(PartialEq)]
enum QuoteMethod {
    Single,
    Double,
    Backtick
}

// Normalised empty space in codeblocks, ensuring that all whitespace in codeblocks is one character wide
pub fn clear_spaces(raw: &str) -> String {
    clear_in_code(&mut raw.chars())
}

fn clear_in_code<'a, 'b>(raw: &mut Chars<'a>) -> String {
    let mut cleared = String::new();
    let mut last_char = 'a';
    while let Some(c) = raw.next() {
        match &c {
            '"' => {
                cleared.push('"');
                println!("Entering double quoted with current \"{}\"", &cleared);
                let res = clear_in_quotes(raw, QuoteMethod::Double);
                cleared.push_str(&res);
                last_char = '"';
            }
            '\'' => {
                cleared.push('\'');
                println!("Entering single quoted with current \"{}\"", &cleared);
                let res = clear_in_quotes(raw, QuoteMethod::Single);
                cleared.push_str(&res);
                last_char = '\'';
            }
            '`' => {
                cleared.push('`');
                println!("Entering backtick quoted with current \"{}\"", &cleared);
                let res = clear_in_quotes(raw, QuoteMethod::Backtick);
                cleared.push_str(&res);
                last_char = '`';
            }
            '{' => {
                cleared.push('{');
                println!("Entering codeblock with current \"{}\"", &cleared);
                let res = clear_in_code(raw);
                cleared.push_str(&res);
                last_char = '}';
            }
            '}' => {
                cleared.push('}');
                println!("Exiting codeblock with current \"{}\"", &cleared);
                return cleared
            }
            x => {
                if last_char.is_whitespace() && x.is_whitespace() {
                    last_char = x.clone();
                    continue;
                }
                cleared.push(x.clone());
                last_char = x.clone();
            }
        }
    };

    cleared
}

fn clear_in_quotes<'a, 'b>(raw: &mut Chars<'a>, method: QuoteMethod) -> String {
    let mut cleared = String::new();
    let mut last_char = 'a';
    while let Some(c) = raw.next() {
        match c {
            '"' => {
                cleared.push('"');
                if method == QuoteMethod::Double && last_char != '\\'  {
                    println!("Exiting double quoted with current \"{}\"", &cleared);
                    return cleared
                }
                last_char = '"';
            }
            '\'' => {
                cleared.push('\'');
                if method == QuoteMethod::Single && last_char != '\\'  {
                    println!("Exiting single quoted with current \"{}\"", &cleared);
                    return cleared
                }
                last_char = '\'';
            }
            '`' => {
                cleared.push('`');
                if method == QuoteMethod::Backtick && last_char != '\\' {
                    println!("Exiting backtick quoted with current \"{}\"", &cleared);
                    return cleared
                }
                last_char = '`';
            }
            '{' => {
                cleared.push('{');
                // Code blocks are only run inside backtick strings and with a $ before the curly braces
                if method == QuoteMethod::Backtick && last_char == '$' {
                    println!("Entering codeblock with current \"{}\"", &cleared);
                    let res = clear_in_code(raw);
                    cleared.push_str(&res);
                    last_char = '}';
                } else {
                    last_char = '{';
                }
            }
            x => {
                cleared.push(x.clone());
                last_char = x.clone();
            }
        }
    }
    cleared
}

fn old_clear_spaces(raw: &str) -> String {
    let mut cleared = String::new();

    let mut currently_quoted = false;
    let mut quote_is_single = false;
    let mut inside_quoted_codeblock = false;
    let mut previous_char = 'a';
    

    for c in raw.chars() {
        if currently_quoted {
            cleared.write_char(c.clone()).expect("Failed to write character to out string during whitespace normalisation");
            previous_char = c;
        } else {
            match c {
                '"' => {
                    if currently_quoted && !quote_is_single && !inside_quoted_codeblock {
                        currently_quoted = false;
                    } else if !currently_quoted {
                        currently_quoted = true;
                        quote_is_single = false;
                        // Just overwrite it here already in case of some weird stuff
                        // Assume that only ever one codeblock will be nested in a string
                        // If there's more, this code will explode
                        // Not really, but spaces might get messed up
                        inside_quoted_codeblock = false;
                    }
                    cleared.write_char(c.clone()).expect("Failed to write character to out string during whitespace normalisation");
                }
                '\'' => {
                    if currently_quoted && quote_is_single {
                        currently_quoted = false;
                    } else {
                        currently_quoted = true;
                        quote_is_single = true;
                    }
                    cleared.write_char(c.clone()).expect("Failed to write character to out string during whitespace normalisation");
                }
                '{' => {
                    if currently_quoted && !inside_quoted_codeblock && !quote_is_single {
                        inside_quoted_codeblock = true;
                    }
                    cleared.write_char(c.clone()).expect("Failed to write character to out string during whitespace normalisation");
                }
                '}' => {
                    if currently_quoted && !inside_quoted_codeblock && !quote_is_single {
                        inside_quoted_codeblock = false;
                    }
                    cleared.write_char(c.clone()).expect("Failed to write character to out string during whitespace normalisation");
                }
                x => {
                    if !(previous_char.is_whitespace() && x.is_whitespace()) {
                        cleared.write_char(x.clone()).expect("Failed to write character to out string during whitespace normalisation");
                    } else {
                    }
                    previous_char = x;
                    
                }
            }
        }
    }

    cleared
}

// Test to make sure that multiple spaces get grouped into one
#[test]
fn test_clear_spaces1() {
    assert_eq!(clear_spaces("  \n  \t"), " ".to_owned())
}

// Test to make sure that only spaces are affected and not normal text
#[test]
fn test_clear_spaces2() {
    let res = clear_spaces("foo  bar");
    assert_eq!(res, "foo bar".to_owned())
}

// Test to make sure that double quoted test keeps spaces
#[test]
fn test_clear_spaces3() {
    let res = clear_spaces("\"foo  bar\"");
    assert_eq!(res, "\"foo  bar\"".to_owned())
}

// Same as previous, but for single quotes
#[test]
fn test_clear_spaces4() {
    let res = clear_spaces("'foo  bar'");
    assert_eq!(res,  "'foo  bar'".to_owned())
}

// Test to make sure that code blocks inside single quotes are treated as quoted as well
#[test]
fn test_clear_spaces5() {
    let res = clear_spaces("pre quote'foo {shouldn\\'t   compress space} inside'post quote");
    assert_eq!(res,  "pre quote'foo {shouldn\\'t   compress space} inside'post quote".to_owned())
}

// Test to make sure that code blocks inside double quotes are treated as non-quoted
#[test]
fn test_clear_spaces6() {
    let res = clear_spaces("pre quote`foo ${should   compress space} inside`post quote");
    assert_eq!(res,  "pre quote`foo ${should compress space} inside`post quote".to_owned())
}